/*=========================================================================================

msprop_cross.c version 1.0

This file is part of the msprop source code and the
examples that accompanies the manuscript:

"A Maxwell-Schrödinger solver for quantum optical few-level systems"

Authors:	Robert Fleischhaker and Jörg Evers

Program:	Robert Fleischhaker

Address:	Max-Planck-Institut für Kernphysik
		Saupfercheckweg 1
		D-69117 Heidelberg
		Germany

Email:		robert.fleischhaker@mpi-hd.mpg.de
		joerg.evers@mpi-hd.mpg.de

=========================================================================================*/
/*********************************************************************
	
	Numerical Integration of Maxwell-Schrödinger equations
	
	msprop_cross: This code is for
	  Example 3: 	5-level system together with chiral SVEA wave equation
			from R. Fleischhaker and J.Evers Phys. Rev. A 80, 063816 (2009)
	
	With local field correction for high density:
	Omega_21 = Omega_21_ex + alpha^2 L rho_21,  L := N lambda^3 / 4 pi^2
	Omega_34 = Omega_34_ex + L rho_34
	Omega_32 = Omega_32_ex + L rho_32
	
	
	Notation: 
	R=[rho21,rho31,rho32,rho34,rho42]
	O=[Omega34,Omega21,Omega32]
	
	Dimensionless quantities:
	z = z` * vg * Tsc
	t = t` * Tsc
	
	Algorithm:
	Adams-Predictor order two for time step of R and
	Lax-Wendroff scheme (order two in space and time) for space step of fields

	Data structur:
	R[0..num_t_steps][0..1][0..dim_R-1] (time coord/last value,one before last/rho_11,rho12,...)
	O[0..num_t_steps][0..1][0..dim_O-1] (time coord/last value,one before last/Omega31,Omega32)
	
	February 2010, Robert Fleischhaker
	
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <complex.h>

//********************************
// Dimension of R and O
//********************************
const int dim_R = 5;
const int dim_O = 3;

//********************************
// Order of Adams-Predictor
// and Lax-Wendroff scheme
//********************************
const int dim_ap = 2;
const int dim_lw = 2;

//********************************
// Parameters:
// 5 x linewidth
// collisional dec
// detuning,
// field strength,
// multiphotonphase,
// helicity
// darkstate parameters
// pulse peak time,
// control field switch on/off time,
// pulse width,
// control field switch on/off width
// group velocity divided by c
// Local field strength
// dimensionless coupling
//********************************

double gam31;
double gam32;
double gam34;
double gam3all;
double gam21;
double gam42;
double gamcol;

double Delta21;

double E34;
double E21;
double E32;

double phimul;

double hel;

double rho011;
double rho044;
double rho041;

double t1;
double t2;
double t3;
double sig1;
double sig2;
double sig3;

int ptype;

double u;
double L;

double eta;
double etaik[3];

const double alpha = 0.00729927; //fine structure constant
const double omegabygamma = 53337724.95755517827; // Sodium 65741897.31115639997; //Rubidium  

//********************************
// Boundary conditions:
// density matrix at t=0
//********************************
complex rho_start[5]={0.,0.,0.,0.,0.};

//********************************
// Parameters for numerics
//********************************

double t_max;
double z_max;

int num_t_steps;
int num_z_steps;

double delta_t;
double delta_z;

int num_t_steps_data;
int num_z_steps_data;

double delta_t_data;
double delta_z_data;


//********************************
// Variables
//********************************

// Density matrix and fields
complex ***R;
complex ***O;

// Time derivatives for Adams-Precdictor
complex **dRdt;

// Parameter and output file
FILE *parfile;
FILE **outputfile;
char parameter_filename[100], output_filename[100];

// Time step index
int i_t;

// Space step index
int i_z;

// Component index
int i_k;



void read_parameters(FILE *parfile)
{
	char input_buffer[200];
	
	// Read titel line
	fgets(input_buffer, 200, parfile);
	
	// Read comment and values
	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	gam31 = strtod(input_buffer, NULL);
	fgets(input_buffer, 200, parfile);
	gam32 = strtod(input_buffer, NULL);
	fgets(input_buffer, 200, parfile);
	gam34 = strtod(input_buffer, NULL);
	fgets(input_buffer, 200, parfile);
	gam21 = strtod(input_buffer, NULL);
	fgets(input_buffer, 200, parfile);
	gam42 = strtod(input_buffer, NULL);
	fgets(input_buffer, 200, parfile);
	gamcol = strtod(input_buffer, NULL);
	
	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	Delta21 = strtod(input_buffer, NULL);
 
	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	E34 = strtod(input_buffer, NULL);
	fgets(input_buffer, 200, parfile);
	E21 = strtod(input_buffer, NULL);
	fgets(input_buffer, 200, parfile);
	E32 = strtod(input_buffer, NULL);
	
	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	phimul = strtod(input_buffer, NULL);

	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	hel = strtod(input_buffer, NULL);

	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	rho011 = strtod(input_buffer, NULL);
	fgets(input_buffer, 200, parfile);
	rho044 = strtod(input_buffer, NULL);
	fgets(input_buffer, 200, parfile);
	rho041 = strtod(input_buffer, NULL);
	
	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	t1 = strtod(input_buffer, NULL);
	fgets(input_buffer, 200, parfile);
	t2 = strtod(input_buffer, NULL);
	fgets(input_buffer, 200, parfile);
	t3 = strtod(input_buffer, NULL);
	
	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	sig1 = strtod(input_buffer, NULL);
	fgets(input_buffer, 200, parfile);
	sig2 = strtod(input_buffer, NULL);
	fgets(input_buffer, 200, parfile);
	sig3 = strtod(input_buffer, NULL);
	
	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	ptype = strtol(input_buffer, NULL, 10);
	
	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	L = strtod(input_buffer, NULL);
	
	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	t_max = strtod(input_buffer, NULL);
 
	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	z_max = strtod(input_buffer, NULL);
 
	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	delta_t = strtod(input_buffer, NULL);
 
	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	delta_z = strtod(input_buffer, NULL);
 
	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	num_t_steps_data = strtol(input_buffer, NULL, 10);
 
	fgets(input_buffer, 200, parfile);
	fgets(input_buffer, 200, parfile);
	num_z_steps_data = strtol(input_buffer, NULL, 10);
}

// Calculate phase from real and imaginary part
// gives values from -pi to pi
double phase(const complex cvalue)
{
	if(creal(cvalue) > 0.)
	{
		return atan(cimag(cvalue) / creal(cvalue));
	}
	if(creal(cvalue) < 0.)
	{
		if(cimag(cvalue) >= 0.)
		{
			return M_PI + atan(cimag(cvalue) / creal(cvalue));
		}
		// cimag(cvalue) < 0.
		return -M_PI + atan(cimag(cvalue) / creal(cvalue));
	}
	// creal(cvalue) == 0.
	if(cimag(cvalue) > 0.)
	{
		return M_PI / 2.;
	}
	if(cimag(cvalue) < 0.)
	{
		return -M_PI / 2.;
	}
	return 0.;
}

// pulse envelope
double pulse_env(const double t, const double t_peak, const double sig, const int ptype)
{
	if(ptype == 1) // Gaussian
		return exp(-(t - t_peak) * (t - t_peak) / 2. / sig / sig);
	else if(ptype == 2) // Sech
		return 1. / cosh((t - t_peak) / sig);
	else if(ptype == 3) // Rectangular
		return 1. / (1. + exp(-(t - (t_peak - sig / 2.)) / sig * 100.)) / (1. + exp((t - (t_peak + sig / 2.)) / sig * 100.));
	else
		return 0.;
}

// control field envelope
double control_env (const double t, const double t_switch, const double sig)
{
	return 1. / (1. + exp(-(t - t_switch) / sig));
}


// initialize parameters
void initialize(void)
{
	read_parameters(parfile);
	fclose (parfile);
	
	num_t_steps = (int) floor(t_max / delta_t + 0.5);
	num_z_steps = (int) floor(z_max / delta_z + 0.5);
	delta_t = t_max / num_t_steps;
	delta_z = z_max / num_z_steps;
	delta_t_data = t_max / num_t_steps_data;
	delta_z_data = z_max / num_z_steps_data;
	
	gam3all = gam31 + gam32 + gam34;
	
	u = 1. / (1. + 3. * omegabygamma * L * gam3all * gam34 / E32 / E32 * rho044);
	eta = 3. / 2. * omegabygamma * L * u * gam3all * gam34;
	printf("u = % .16f\n", u);
	printf("eta = % .16f\n", alpha);
	if(u * delta_z > delta_t)
	{
		printf("u * delta_z > delta_t\n");
		printf("delta_z: % .16f\n", delta_z);
		printf("delta_t: % .16f\n", delta_t);
		exit(1);
	}
	
	etaik[0] = eta;
	etaik[1] = eta * alpha;
	etaik[2] = eta * gam32 / gam34;

	// Allocate memory, all in one block
	// Timegrid
	R = (complex ***) malloc((num_t_steps + 1) * sizeof(complex **));
	O = (complex ***) malloc((num_t_steps + 1) * sizeof(complex **));
	// Different time steps
	R[0] = (complex **) malloc((num_t_steps + 1) * dim_lw * sizeof(complex *));
	O[0] = (complex **) malloc((num_t_steps + 1) * dim_lw * sizeof(complex *));
	dRdt = (complex **) malloc(dim_ap * sizeof(complex *));
	// Different variables
	R[0][0] = (complex *) calloc((num_t_steps + 1) * dim_lw * dim_R, sizeof(complex));
	O[0][0] = (complex *) calloc((num_t_steps + 1) * dim_lw * dim_O, sizeof(complex));
	dRdt[0] = (complex *) calloc(dim_ap * dim_R, sizeof(complex));
	// Assign address values
	for (i_t = 0; i_t <= num_t_steps; i_t++)
	{
		R[i_t] = &R[0][dim_lw * i_t];
		O[i_t] = &O[0][dim_lw * i_t];
		for(i_z = 0; i_z < dim_lw; i_z++)
		{
			R[i_t][i_z] = &R[0][0][dim_lw * dim_R * i_t + dim_R * i_z];
			O[i_t][i_z] = &O[0][0][dim_lw * dim_O * i_t + dim_O * i_z];
		}
	}
	for(i_t = 0; i_t < dim_ap; i_t++)
		dRdt[i_t] = &dRdt[0][dim_R * i_t];

/*	//Test of data structure
	for(i_t = 0; i_t < (num_z_steps + 1) * dim_ap * dim_O; i_t++)
		O[0][0][i_t] = i_t;
	for(i_z = 0; i_z <= num_z_steps; i_z++)
		for(i_t = 0; i_t < dim_ap; i_t++)
			for(i_k = 0; i_k < dim_O; i_k++)
				printf("i_z: %i \ti_t: %i \ti_k: %i\tO: %.1f\n",i_z ,i_t, i_k, cabs(O[i_z][i_t][i_k]));*/
	
	// prepare output files
	int data_it;
	char *filename;
	filename = (char *) malloc(30 * sizeof (char));
	outputfile = (FILE **) malloc((num_t_steps_data + 1) * sizeof(FILE *));
	for(data_it = 0; data_it <= num_t_steps_data; data_it++)
	{
		sprintf(filename, "data/%s_t=%.2f.dat", output_filename, data_it * delta_t_data);
		outputfile[data_it] = fopen(filename, "w");
		if(outputfile[data_it] == NULL)
		{
			printf("Cannot open outputfile %s!", filename);
			exit(EXIT_FAILURE);
		}
	}
	free(filename);
	
	// Boundary condition
	for(i_t = 0; i_t <= num_t_steps; i_t++)
	{
		O[i_t][0][0] = E34 * pulse_env(i_t * delta_t, t1, sig1, ptype);
		O[i_t][0][1] = -hel * I * alpha* E21 * pulse_env(i_t * delta_t, t1, sig1, ptype);
		O[i_t][0][2] = E32 * cexp(I * phimul * (control_env(i_t * delta_t, t2, sig2) + control_env(i_t * delta_t, t3, sig3)));
	}
}

// RHS of atomic ODE
void rhs_dRdt(const complex R[], const complex O[], const double z, const double t, complex dRdt[])
{
	// d/dt\[Rho]21
	dRdt[0] = -(gam21*R[0])/2. + 0.5*I*alpha*alpha*L*rho011*gam34*R[0] - gamcol*R[0] + I*Delta21*R[0] + 0.5*I*conj(L*gam34*R[2] + O[2])*R[1] + 0.5*I*rho011*O[1];
	
	// d/dt\[Rho]31
	dRdt[1] = -(gam3all*R[1])/2. - 2*gamcol*R[1] + I*Delta21*R[1] + 0.5*I*L*gam34*R[0]*R[2] - 0.5*I*alpha*alpha*L*gam34*R[0]*R[2] + 0.5*I*L*rho041*gam34*R[3] - 0.5*I*R[2]*O[1] + 0.5*I*R[0]*O[2] + 0.5*I*rho041*O[0];
	
	// d/dt\[Rho]32
	dRdt[2] = -0.5*I*conj(alpha*alpha*L*gam34*R[0] + O[1])*R[1] - (gam21*R[2])/2. - (gam3all*R[2])/2. - gamcol*R[2] + 0.5*I*L*gam34*R[3]*R[4] + 0.5*I*R[4]*O[0];
	
	// d/dt\[Rho]34
	dRdt[3] = 0.5*I*L*conj(R[4])*gam34*R[2] + 0.5*I*L*rho044*gam34*R[3] - (gam3all*R[3])/2. - gamcol*R[3] + I*Delta21*R[3] + 0.5*I*conj(R[4])*O[2] + 0.5*I*rho044*O[0];
	
	// d/dt\[Rho]42
	dRdt[4] = -0.5*I*rho041*conj(alpha*alpha*L*gam34*R[0] + O[1]) + 0.5*I*conj(L*gam34*R[3] + O[0])*R[2] - 0.5*I*L*conj(R[3])*gam34*R[2] - (gam21*R[4])/2. - gam42*R[4] - I*Delta21*R[4] - 0.5*I*conj(R[3])*O[2];
}

// Source function for wave equations
complex Rp(const int i_t, const int i_z, const int i_k)
{
	switch(i_k)
	{
		case 0:
			return R[i_t][i_z][3] + I * hel * alpha * R[i_t][i_z][0];
		case 1:
			return alpha * R[i_t][i_z][0] - I * hel * R[i_t][i_z][3];
		case 3:
			return R[i_t][i_z][2];
	}
	return (complex)0;
}

// Lax-Wendroff scheme for first space step
complex tlw_fss_O_new(const int i_k)
{
	return O[0][0][i_k] + u * delta_z / delta_t / 2. * (
			O[0][1][i_k] - O[1][0][i_k] + 
			u * delta_z / delta_t * (O[1][0][i_k] - O[0][0][i_k]) + 
			I * etaik[i_k] * delta_z / 2. * (Rp(0, 1, i_k) - Rp(1, 0, i_k))) + 
			I * etaik[i_k] * delta_z / 2. * (Rp(0, 0, i_k) + Rp(0, 1, i_k));
}

//usual Lax-Wendroff scheme
complex tlw_O_new(const int i_k, const int i_t)
{
	return O[i_t][0][i_k] + u * delta_z / delta_t / 2. * (
			O[i_t - 1][1][i_k] - O[i_t + 1][0][i_k] + 
			u * delta_z / delta_t * (O[i_t + 1][0][i_k] - 2. * O[i_t][0][i_k] + O[i_t - 1][1][i_k]) + 
			I * etaik[i_k] * delta_z / 2. * (Rp(i_t - 1, 1, i_k) - Rp(i_t + 1, 0, i_k))) + 
			I * etaik[i_k] * delta_z / 2. * (Rp(i_t, 0, i_k) + Rp(i_t, 1, i_k));
}

// Lax-Wendroff for last space step
complex tlw_lss_O_new(const int i_k)
{
	return 1. / (1. + u * delta_z / delta_t / 2.) * (
			O[num_t_steps][0][i_k] + u * delta_z / delta_t / 2. * 
			(O[num_t_steps - 1][0][i_k] + O[num_t_steps - 1][1][i_k] - O[num_t_steps][0][i_k]) + 
			I * etaik[i_k] * delta_z / 2. * (Rp(num_t_steps, 1, i_k) + Rp(num_t_steps, 0, i_k)));
}

// Export time line into output files
void write_timeline(double z)
{
	// time step index
	int i_t;
	
	printf("Write z = % .2f ...", z);
	
	int data_it;
	double pulsephase;
	
	// Write solution to files
	for (data_it = 0; data_it <= num_t_steps_data; data_it++)
	{
		i_t = floor(data_it * delta_t_data / delta_t + 0.5);
		
		if(cabs(O[i_t][0][0] / E34) >= 0.01)
			pulsephase = phase(O[i_t][0][0]);
		else
			pulsephase = 0.0;
		
		fprintf (outputfile[data_it], "% .16f\t% .16f\t% .16f\t% .16f\t% .16f\t% .16f\t% .16f\t% .16f\t% .16f\t% .16f\t% .16f\t% .16f\t% .16f\t% .16f\n",
				 z,
				 cabs(O[i_t][0][0] / E34),
				 cabs(O[i_t][0][1] / E21 / alpha),
				 cabs(O[i_t][0][2] / E32),
				 pulsephase,
				 phase(O[i_t][0][1]),
				 phase(O[i_t][0][2]),
				 cabs(R[i_t][0][0]),
				 cabs(R[i_t][0][1]),
				 cabs(R[i_t][0][2]),
				 cabs(R[i_t][0][3]),
				 cabs(R[i_t][0][4]),
				 creal(O[i_t][0][0] / E34),
				 cimag(O[i_t][0][0] / E34));
	}
	printf("done\n");
}


void finish(void)
{
	// free memory
	free(R[0][0]);
	free(R[0]);
	free(R);
	free(O[0][0]);
	free(O[0]);
	free(O);
	
	int data_it;
	for(data_it = 0; data_it <= num_t_steps_data; data_it++)
		fclose(outputfile[data_it]);
	free(outputfile);
}


//**********************************
// Main program
//**********************************
int main(int argc, char *argv[])
{
	// Output interval
	double next_data_z = 0.;
	
	// Pointer to swap addresses of last and newest point
	complex *swap;	
	
	printf("\n\n Pulse propagation dynamics\n\n");
 
	//**********************************
	// Read parameter
	//**********************************
	if (argc != 3) 
	{
		printf ("Wrong number of parameters! \n\n");
		printf ("usage: executable-file parameter-file output-file\n\n");
		exit(1);
	}
	strcpy (parameter_filename, argv[1]);
	strcpy (output_filename, argv[2]);
	
	// Make dir
	if(system("test -d data"))
		system("mkdir data");
	else
	{
		system("rm -r data");
		system("mkdir data");
	}
	
	parfile = fopen(parameter_filename, "r");
 
	if(parfile == NULL)
	{
		printf("Cannot open parameter file '");
		printf(parameter_filename);
		printf("'!\n");
		exit(1);
	}
	
	initialize();
	
	
	// start numerical integration
	
	// integrate initial time line
	for(i_k = 0; i_k < dim_R; i_k++)
		R[0][0][i_k] = rho_start[i_k];
	for(i_t = 0; i_t < num_t_steps; i_t++)
	{
		// Swap adresses of memory of new and last derivative
		dRdt[1] = dRdt[0];
		// Derivatives at newest point
		rhs_dRdt(R[i_t][0], O[i_t][0], 0., i_t * delta_t, dRdt[0]);
		// Put calculated point into memory
		for(i_k = 0; i_k < dim_R; i_k++)
			R[i_t + 1][0][i_k] = R[i_t][0][i_k] + delta_t / 2. * (3. * dRdt[0][i_k] - dRdt[1][i_k]);
	}
	
	// Write initial time line
	write_timeline(0.);
	next_data_z += delta_z_data;
	
	// Main evolution loop
	for (i_z = 0; i_z < num_z_steps; i_z++)
	{
		// all other time lines
		
		// initial conditions for R
		for(i_k = 0; i_k < dim_R; i_k++)
		{
			R[0][1][i_k] = rho_start[i_k];
			dRdt[0][i_k] = 0.;
		}
		// Swap adresses of memory of new and last point
		swap = R[0][0];
		R[0][0] = R[0][1];
		R[0][1] = swap;
		
		// first space step
		for(i_k = 0; i_k < dim_O; i_k++)
			O[0][1][i_k] = 0.0; //tlw_fss_O_new(i_k);
		// Swap adresses of memory of new and last point
		swap = O[0][0];
		O[0][0] = O[0][1];
		O[0][1] = swap;
		
		// first time step
		// Swap adresses of memory of new and last derivative
		dRdt[1] = dRdt[0];
		// Derivatives at newest point
		rhs_dRdt(R[0][0], O[0][0], (i_z + 1) * delta_z, 0., dRdt[0]);
		// Put calculated point into memory
		for(i_k = 0; i_k < dim_R; i_k++)
			R[1][1][i_k] = R[0][0][i_k] + delta_t / 2. * (3. * dRdt[0][i_k] - dRdt[1][i_k]);
		// Swap adresses of memory of new and last point;
		swap = R[1][0];
		R[1][0] = R[1][1];
		R[1][1] = swap;
		
		// Intermediate points
		// Do space step for O and time step for R alternately
		for(i_t = 1; i_t < num_t_steps; i_t++)
		{
			// Calculate next O value with Lax-Wendroff scheme
			for(i_k = 0; i_k < dim_O; i_k++)
				O[i_t][1][i_k] = tlw_O_new(i_k, i_t);
			// Swap adresses of memory of new and last point
			swap = O[i_t][0];
			O[i_t][0] = O[i_t][1];
			O[i_t][1] = swap;
			
			// Calculate next R value with adams-predictor of order two
			// Swap adresses of memory of new and last derivative
			dRdt[1] = dRdt[0];
			// Derivatives at newest point
			rhs_dRdt(R[i_t][0], O[i_t][0], (i_z + 1) * delta_z, i_t * delta_t, dRdt[0]);
			// Put calculated point into memory
			for(i_k = 0; i_k < dim_R; i_k++)
				R[i_t + 1][1][i_k] = R[i_t][0][i_k] + delta_t / 2. * (3. * dRdt[0][i_k] - dRdt[1][i_k]);
			// Swap adresses of memory of new and last point;
			swap = R[i_t + 1][0];
			R[i_t + 1][0] = R[i_t + 1][1];
			R[i_t + 1][1] = swap;
		}
	
		// Calculate last O value with last space step Lax-Wendroff scheme
		for(i_k = 0; i_k < dim_O; i_k++)
			O[num_t_steps][1][i_k] = tlw_lss_O_new(i_k);
		// Swap adresses of memory of new and last point
		swap = O[num_t_steps][0];
		O[num_t_steps][0] = O[num_t_steps][1];
		O[num_t_steps][1] = swap;
		
		// write to file if data z value
		if(fabs((i_z + 1) * delta_z - next_data_z) <= delta_z / 2.)
		{
			write_timeline((i_z + 1) * delta_z);
			next_data_z += delta_z_data;
		}
		
	}
	
	FILE * timelinefile = fopen("timeline.dat", "w");
	if(timelinefile == NULL)
	{
		printf("Cannot open timeline.dat!");
		exit(EXIT_FAILURE);
	}
	
	int data_it;
	double pulsephase;
	for (data_it = 0; data_it <= num_t_steps_data; data_it++)
	{
		i_t = floor(data_it * delta_t_data / delta_t + 0.5);
		if(cabs(O[i_t][0][0] / E34) >= 0.01)
			pulsephase = phase(O[i_t][0][0]);
		else
			pulsephase = 0.0;
		fprintf(timelinefile,"% .16f\t% .16f\t% .16f\t% .16f\t% .16f\t% .16f\t% .16f\n",
		i_t * delta_t,
		creal(O[i_t][0][0] / E34),
		cimag(O[i_t][0][0] / E34),
		cabs(O[i_t][0][0] / E34),
		pulsephase,
		cabs(O[i_t][0][2] / E32),
		phase(O[i_t][0][2]));
	}
	fclose(timelinefile);
	
	finish();
	return 0;
}
